;; Pegasus Mbox converter
;  This program takes the file format Pegasus mail uses when
;  it saves messages as one file and converts it to a working
;  mbox file. Pretty simple really.
;
;  Copyright (c) 2005
;       Aaron Hsu
;
;  Redistribution and use in source and binary forms, with or without
;  modification, are permitted provided that the following conditions
;  are met:
;  1. Redistributions of source code must retain the above copyright
;     notice, this list of conditions and the following disclaimer.
;  2. Redistributions in binary form must reproduce the above copyright
;     notice, this list of conditions and the following disclaimer in the
;     documentation and/or other materials provided with the distribution.
;  3. Neither the name of Aaron Hsu nor the names of other contributors
;     may be used to endorse or promote products derived from this software
;     without specific prior written permission.
; 
;  THIS SOFTWARE IS PROVIDED BY AARON HSU AND CONTRIBUTORS ``AS IS'' AND
;  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  SUCH DAMAGE.
;
;  The basic logic of the program is to read each line of the given file,
;  until either "-- End --" or "Return-Path" is found at the start of one
;  of the lines. It will make sure that the requisite "From " is not already
;  there, and if it is not, it will insert the right one by reading ahead
;  and finding the right "From: " Header.
;
;; convert-pegasus-file : filename(string) outputfile(string) -> undefined
;  loops through a given pegasus "export mail as a single file" file and
;  removes the added "-- End --" delimiter, replacing that line with the
;  correct mbox formatted delimiters, which makes the message a valid and
;  readable mbox file.
(define (convert-pegasus-file filename outputname)
  (require 'line-i/o) ; Require line-i/o from SLIB
  (require 'regex)
  (let* ((ifile (open-file filename open_read)) ;Open for Reading Only
         (ofile (open-file outputname open_write)) ;Open for Writing Only
         ;; find-name : From:(string)->email(string)
         ;  Takes a From: header from an email address in the form of a
         ;  single string, and exstracts the email address from the header.
         (extract-name
           (lambda (cline)
             (let ((regs (regsearch " <.*@.*>$" cline))) ; Discover Format
               (substring ; We need the substring
                 cline    ; of the Date: header
                 (if regs ; Name + Address Format
                     (+ regs 2) ; Addition compensates for " <"
                     7) ; just an email address (no name) so skip "From: "
                 (if regs
                     (- (string-length cline) 1) ; Take out ">" unless,
                     (string-length cline)))))) ; there is No ">"
         ;; find-date : Date-header(string)->Date(string)
         ;  Takes the Date: header from an email address as a single string
         ;  and extracts the date in a string which can be used within the
         ;  "From " section of the mbox format
         (extract-date 
           (lambda (cline)
             (let* ((datestrg (substring 
                                cline
                                0
                                (+ (regsearch ; Cut out last pieces of Date
                                     " [[:digit:]]*:" cline) 9)))
                    (clen (string-length datestrg))
                    (fldiff (if (string-index datestrg #\,) ; Day exists
                                0  
                                5))) ; Day does not exist in Date line
               (string-append ; Composite Date string
                 (substring datestrg (- 6 fldiff) (- 9 fldiff)) ;Day of Week
                 " "
                 (substring datestrg (- clen 17) (- clen 14)) ;Month
                 " "
                 (substring datestrg (- 11 fldiff) (- clen 18)) ; Day
                 " "
                 (substring datestrg (- clen 8) clen) ;Time
                 " "
                 (substring datestrg (- clen 13) (- clen 9)))))) ; Year
         ;; get-header : input-port extract-function regexp-string
         ;            -> header-buffer(list)
         ;  get-header is used to read ahead in a file and obtain information
         ;  from a header which has not yet been read. get-header reads
         ;  through the file a line at a time, and runs the extract-header
         ;  function on that line, assuming that function will return a
         ;  string, which matches the regular expression string given in
         ;  re-string. It then returns that extracted string as the first
         ;  element of the returned list, with all other elements being the
         ;  newly read lines in this function, including the header line
         ;  which matched re-string.
         (get-header
           (lambda (ifile extract-header re-string)
             (let loop ((buf (list))
                        (cline (read-line ifile)))
               (if (regmatch? re-string cline)
                   (append
                     (list (extract-header cline))
                     buf
                     (list cline))
                   (loop (append buf (list cline)) (read-line ifile))))))
         ;; ra-from : date-buffer(list) current-line(string) buffer(list)
         ;         -> new-buffer(list)
         ;  ra-from composes a "From " mbox message delimiter assuming that
         ;  the current line passed as a string contains the "From: " header
         ;  and that the d variable is a date-buffer returned from get-header
         ;  with the date information as the first element of the list, and
         ;  the rest as the buffer which was read during the last get-header
         ;  process.
         (ra-from
           (lambda (d cline buf)
             (append
               (list (string-append
                       "From "
                       (extract-name cline) ; find proper email addy
                       " "
                       (car d))) ; date has already been extracted
               buf
               (list cline)
               (cdr d))))
         ; ra-date is the same format as ra-from, but it assumes that the
         ; from header has already been extracted by get-header instead of
         ; the date header. It assumes that the date header is cline.
         (ra-date
           (lambda (f cline buf)
             (append
               (list (string-append
                       "From "
                       (car f) ; email addy already extracted
                       " "
                       (extract-date cline))) ; extract Date
               buf
               (list cline)
               (cdr f))))
         ;; read-ahead : input-port line-buffer(list) -> line-buffer(list)
         ;  reads ahead in the file to find the correct information to
         ;  compose an mbox style message delimiter, and returns a modified
         ;  list with this delimiter at the head of the lines read. The
         ;  buffer format is a list element for each line, and each line
         ;  represented by a string.
         (read-ahead
           (lambda (ifile buf)
             (let loop ((buf buf)
                        (curline (read-line ifile)))
               (if (not (eof-object? curline))
                   (cond ((regmatch? "^From " curline) ; delimiter exists
                          (append buf (list curline))) ; don't do anything
                         ((regmatch? "^From: " curline) ; found from first
                          (ra-from
                            (get-header ifile extract-date "^Date: ")
                            curline
                            buf))
                         ((regmatch? "^Date: " curline) ; found date first
                          (ra-date
                            (get-header ifile extract-from "^From: ")
                            curline
                            buf))
                         (else
                           (loop
                             (append buf (list curline))
                             (read-line ifile))))
                   buf))))
         ;; write-buf : line-buffer(list) output-port -> undefined
         ;  writes a line buffer (a list of strings representing lines) to
         ;  a given output port, assuming the first element of the list is
         ;  the first to be written.
         (write-buf
           (lambda (buf ofile)
             (let loop ((buf buf))
               (cond ((not (null? buf))
                      (display (car buf) ofile)
                      (newline ofile)
                      (loop (cdr buf))))))))
    (let ((buf (list (read-line ifile)))) ; Initial delimiter
      (write-buf (read-ahead ifile buf) ofile))
    (let readloop ((buf (list (read-line ifile))))
      (if (not (eof-object? (car buf)))
          (cond ((string=? (car buf) "-- End --")
                 (write-buf (read-ahead ifile (cdr buf)) ofile)
                 (readloop (list (read-line ifile))))
                (else
                  (write-buf buf ofile)
                  (readloop (list (read-line ifile)))))))))
;; convert-pegasus-file-list file-list -> undefined
;  This simple wrapper calls convert-pegasus-file on each element in the
;  given list, using the export name of the list element plus ".new"
(define (convert-pegasus-file-list lst)
  (if (not (null? lst))
      (begin
        (display (car lst))
        (newline)
        (convert-pegasus-file (car lst) (string-append (car lst) ".new"))
        (convert-pegasus-file-list (cdr lst)))))
